/**
 * Copyright (c) Omnissa, LLC. All rights reserved.
 * This product is protected by copyright and intellectual property laws in the
 * United States and other countries as well as by international treaties.
 * -- Omnissa Restricted
 *
 * @format
 */

/**
 * TODO: i18n for context menu(P1 ctritical)
 * TODO: check clientWindow for multi-session support(P1 serious)
 */
//import SDKBackgroundEntry from previous file

const URL_REDIRECTION_EXTENSION_ID = "lfidjngibpklhhijphdmbmedchiiolgk";
const URL_REDIRECTION_RULE_SOURCE_KEY = "URL_REDIRECTION_RULE_SOURCE";

var LAUNCHAPP = function () {
   // Policy Name
   this.policyName = "Horizon Client Policy";
   // Default index page
   this.index = "./webclient/index.html";
   this.fileAssociationWindow = "./webclient/file-association.html";
   this.shelfPage = "./webclient/mock-shelf-window.html";
   /**
    * Allow client reuse for URL launches(extendable to SDK)
    * Only support desktop/app launching for now
    */
   this.activeClient = null;
   this.isFALaunchWithMainWindowHide = false;
   this.fileInfo = {};
   this.clearActiveClient("background initialied reset");

   // Main window reference
   this.mainWindow = null;
   this.shelfPlaceHolderWindow = null;

   // content window API object reference
   this.chromeClient = null;

   // Cache some data which could be used across windows.
   this.cachedObj = {};
   this.ws1ModeEnabled = null;
   this.SDKModeEnabled = null;
   this.launchSettings = null;
   this.skipClearForWaiting = false;
};

var showClientAsSetting = null;

/**
 * use below implementation can reduce code mounted in background page, but introduce small copy of codes from JSCDK.
 * We need to keep below updated if XML API changed
 */
LAUNCHAPP.prototype.logoutActiveClient = async function () {
   if (!this.activeClient || !this.activeClient.xmlServerURL) {
      console.log("no active client need to be logout");
      return;
   }
   if (
      typeof this.activeClient.serverAddress !== "string" ||
      typeof this.activeClient.xmlServerURL !== "string" ||
      !this.activeClient.xmlServerURL.startsWith("https://")
   ) {
      console.error("invalid server address" + this.activeClient.serverAddress + ", " + this.activeClient.xmlServerURL);
      return;
   }
   if (this.skipClearForWaiting) {
      // if there is any other handling trigger the client close on purpose, this default logout handler should not tick in.
      console.log("skip client logout for script triggered close");
      return;
   }
   // this.activeClient.serverAddress can't be directly used, since mid is needed for F5
   let requestAddress = this.activeClient.xmlServerURL;
   // We must keep below updated if XML API changed
   const messageText = "<?xml version='1.0' encoding='UTF-8'?><broker version='15.0'><do-logout/></broker>";

   // use fetch to keep consistent and avoid response handling codes
   // no CSRF token needed for chrome client.
   try {
      console.log("logout client form server: " + this.activeClient.serverAddress);
      await window.fetch(requestAddress, {
         method: "POST",
         mode: "same-origin",
         cache: "no-cache",
         headers: {
            "Content-Type": "text/xml;charset=UTF-8"
         },
         credentials: "same-origin",
         referrerPolicy: "no-referrer",
         keepalive: false,
         body: messageText
      });
      console.log("response received for logout");
   } catch (e) {
      console.error(e);
   }
};
LAUNCHAPP.prototype.clearActiveClient = function (reasonString, logoutSession) {
   if (logoutSession) {
      this.logoutActiveClient();
   }
   if (!this.activeClient || !this.activeClient.client) {
      console.log("no active client to be reset for " + reasonString);
      return;
   } else {
      console.log("resetting active client for " + reasonString);
   }
   this.activeClient = {
      client: null,
      operations: {
         launchItem: null //URL param must contains SAMLart, item info
      },
      identityId: null,
      serverAddress: null,
      isReusable: false,
      xmlServerURL: null,
      userName: null,
      userDomain: null
   };
   if (this.skipClearForWaiting) {
      this.skipClearForWaiting = false;
      console.log("keep waiting for launch, since client clear is expected for launching");
   } else {
      console.log("stop waiting for launching, since reusable client clear detected");
      this.stopWaitForLaunch();
   }
};

// use explicit parmeter for readability in JS
LAUNCHAPP.prototype.setActiveClient = function ({
   isReusable,
   identityId,
   serverAddress,
   xmlServerURL,
   operations,
   client,
   userName,
   userDomain
}) {
   console.log("setting active client");
   if (!client || !serverAddress || !xmlServerURL || typeof isReusable !== "boolean") {
      console.error("invalid active client");
      return false;
   }
   if (isReusable && (!operations || typeof operations.launchItem !== "function")) {
      console.error("invalid reusable client");
      return false;
   }
   this.activeClient = {
      client: client,
      operations: operations,
      identityId: identityId,
      serverAddress: serverAddress,
      xmlServerURL: xmlServerURL,
      isReusable: isReusable,
      userName: userName,
      userDomain: userDomain
   };
   console.log("active client had been set for server: " + serverAddress);
   return true;
};
LAUNCHAPP.prototype.stopWaitForLaunch = function () {
   if (this.sdk) {
      console.log("stop wait and accept next launch request");
      setTimeout(() => {
         this.sdk.onRemoteRequestProcessed();
      });
   }
};
LAUNCHAPP.prototype.isSameHost = function (host1, host2) {
   if (host1 === host2) {
      return true;
   }
   if (typeof host1 === "string" && !host1.includes(":")) {
      host1 += ":443";
   }
   if (typeof host2 === "string" && !host2.includes(":")) {
      host2 += ":443";
   }
   return host1 === host2;
};

LAUNCHAPP.prototype.canReuseClientforWS1 = function (targetInfo) {
   if (!targetInfo || !this.activeClient) {
      return false;
   }
   let hasActiveClient =
      !!targetInfo.horizonId &&
      this.activeClient &&
      this.activeClient.isReusable &&
      this.activeClient.identityId === targetInfo.horizonId;
   let hasAuthInfo = !!targetInfo.SAMLart;
   let sameServer = this.isSameHost(targetInfo.host, this.activeClient.serverAddress);
   let notUseExisting = targetInfo.useExisting === "false";

   let hasTargetpool = !!(
      targetInfo.applicationId ||
      targetInfo.applicationName ||
      targetInfo.desktopId ||
      targetInfo.desktopName ||
      targetInfo.collabSessionId ||
      targetInfo.itemId
   );
   let action = targetInfo.action ? targetInfo.action.toLowerCase() : null;
   let hasUnsupportedAction =
      !!action && (action === "logoff" || action === "reset" || action === "restart" || action === "browse");
   return hasActiveClient && hasAuthInfo && hasTargetpool && !hasUnsupportedAction && sameServer && !notUseExisting;
};

/**
 * TODO: refactor to split into separated file
 *
 * For AD for SDK, we use strict match, without abstract domain from user name.
 * while we treat it as mismatch for default domain to specified domain for security reason.
 * For SAML, we always assume same user, and 3rd party app should be responsible to logout previous by SDK when user switch
 *
 * Thus SDK consuming app need to:
 * - use consistent style of user name and domain, instead using UPN in some case and use domain/user in other cases
 * - use SDK to close client when detect user changed/logout, especially for SAMLart setup
 */
LAUNCHAPP.prototype.isSameUser = function (user1, domain1, user2, domain2) {
   console.log("checking identical user for ", user1, domain1, user2, domain2);
   return (user1 === user2 || (!user1 && !user2)) && (domain1 === domain2 || (!domain1 && !domain2));
};

LAUNCHAPP.prototype.canReuseClientforSDK = function (targetInfo, options) {
   if (!targetInfo || !this.activeClient) {
      return false;
   }
   let generalMatch = !targetInfo.serverId && !!this.activeClient.identityId;
   let strictMatch = !!targetInfo.serverId && this.activeClient.identityId === targetInfo.serverId;

   let hasActiveClient = generalMatch || strictMatch;
   let hasAuthInfo = !!targetInfo.SAMLart || (!!options && !!options.authSecret);
   let sameServer = this.isSameHost(targetInfo.host, this.activeClient.serverAddress);
   let sameUser = this.isSameUser(
      targetInfo.user,
      targetInfo.domainName,
      this.activeClient.userName,
      this.activeClient.userDomain
   );
   console.log("client reuse check for SDK ", { sameServer, sameUser, hasAuthInfo, hasActiveClient });

   let hasTargetpool = !!(
      targetInfo.applicationId ||
      targetInfo.applicationName ||
      targetInfo.desktopId ||
      targetInfo.desktopName ||
      targetInfo.collabSessionId ||
      targetInfo.itemId
   );
   return hasActiveClient && hasAuthInfo && hasTargetpool && sameServer && sameUser;
};

/**
 * When launching the client via URL redirection, check whether the active client can reuse.
 */
LAUNCHAPP.prototype.canReuseClientForUrlRedirection = function (targetInfo) {
   if (!targetInfo || !this.activeClient || !targetInfo.url) {
      return false;
   }

   const params = this.parseURI(targetInfo.url);

   return this.isSameHost(targetInfo.host, this.activeClient.serverAddress) && !!params.url;
};

/**
 * currently only support launch desktop/app
 */
LAUNCHAPP.prototype.doURLAction = function (targetInfo, options, isSDKLaunched) {
   // when it's titan auth call back launched, check whether client can call doTitanLaunch
   if (this.chromeClient && targetInfo.code !== "" && targetInfo.host === "https:") {
      // when cancel the titan launch process, isTitanLaunch is false
      this.showClientMainWindow();
      if (this.chromeClient.isTitanLaunch && typeof this.chromeClient.titanLaunch === "function") {
         this.chromeClient.titanLaunch(targetInfo.url);
      }
      return true;
   }
   // when it's SDK launched, check whether client can reuse by function canReuseClientforSDK
   if (isSDKLaunched === true && this.canReuseClientforSDK(targetInfo, options)) {
      console.log("reuse active client for launch");
      targetInfo.launchType = "SDK";
      this.activeClient.operations.launchItem(targetInfo, options);
      return true;
   }
   // when it's WS1 launched, check whether client can reuse by function canReuseClientforWS1
   if (!!targetInfo.horizonId && this.canReuseClientforWS1(targetInfo)) {
      console.log("reuse active client for launch");
      targetInfo.launchType = "WS1";
      this.activeClient.operations.launchItem(targetInfo, options);
      return true;
   }
   // when it's URL redirection launched, check whether client can reuse by function canReuseClientForUrlRedirection
   if (this.canReuseClientForUrlRedirection(targetInfo)) {
      console.log("reuse active client for launch");
      targetInfo.launchType = "UrlRedirection";
      targetInfo.url = this.parseURI(targetInfo.url).url;
      this.activeClient.operations.launchItem(targetInfo, options);
      return true;
   }

   console.log("no active client to be reused for launch");
   return false;
};

/**
 * below functions are written in multi-session compatible way, even in ECR version
 * we don't support it.
 */
LAUNCHAPP.prototype.setupForURL = function () {
   var self = this;

   /**
    * Mainly used for Error dialog
    */
   self.chromeClient.showClient = () => {
      self.showClientMainWindow();
   };

   /**
    * Mainly used for Error/setting dialog closed
    */
   self.chromeClient.onSettingClosed = self.chromeClient.hideClient = () => {
      self.hideClientMainWindow();
   };

   /**
    * Should only called once, when XML response returned.
    */
   self.chromeClient.onWS1ModeInited = async function (ws1ModeEnabled, hideClientMainWindow = true) {
      console.log("WS1 mode enabled: " + ws1ModeEnabled);
      self.ws1ModeEnabled = ws1ModeEnabled;
      delete self.chromeClient.onWS1ModeInited;
      self.chromeClient.isClientHidden = ws1ModeEnabled;

      if (ws1ModeEnabled) {
         try {
            await self.setupForWs1orSDKMode();
         } catch (e) {
            console.log(e);
         }

         if (hideClientMainWindow) {
            console.log("hide client since WS1 mode detected");
            self.hideClientMainWindow();
         }
      }

      if ((self.ws1ModeEnabled === false && self.SDKModeEnabled === false) || hideClientMainWindow === false) {
         console.log("show client when both SDK and WS1 has no request to hide client or need to auth in webview.");
         self.showClientMainWindow();
      }
   };

   /**
    * Should only called once, when XML response returned.
    */
   self.chromeClient.onSDKModeInited = async function (SDKModeEnabled) {
      console.log("SDK hide mode enabled: " + SDKModeEnabled);
      self.SDKModeEnabled = SDKModeEnabled;
      delete self.chromeClient.onSDKModeInited;
      self.chromeClient.isClientHidden = SDKModeEnabled;

      if (SDKModeEnabled) {
         try {
            await self.setupForWs1orSDKMode();
         } catch (e) {
            console.log(e);
         }
         console.log("hide client since SDK request so");
         self.hideClientMainWindow();
      }

      self.SDKModeEnabled = SDKModeEnabled;
      if (self.ws1ModeEnabled === false && self.SDKModeEnabled === false) {
         console.log("show client both SDK and WS1 has no request to hide client");
         self.showClientMainWindow();
      }
   };

   self.chromeClient.stopWaitForLaunch = self.stopWaitForLaunch.bind(self);
};
LAUNCHAPP.prototype.showClientMainWindow = function () {
   if (this.mainWindow && !this.ws1ModeEnabled && !this.SDKModeEnabled) {
      this.mainWindow.show();
      this.mainWindow.focus();
      return;
   }
   if (!this.mainWindow || !this.shelfPlaceHolderWindow) {
      console.log("show client failed due to window is absent.");
      return;
   }
   this.mainWindow.show();
   this.mainWindow.focus();
   this.shelfPlaceHolderWindow.hide();
   this.chromeClient.isClientHidden = false;
   console.log("show client successfully.");
};
LAUNCHAPP.prototype.hideClientMainWindow = function () {
   if (!this.mainWindow || !this.shelfPlaceHolderWindow) {
      console.log("hide client failed due to window is absent.");
      return;
   }
   this.mainWindow.hide();
   this.shelfPlaceHolderWindow.show();
   this.chromeClient.isClientHidden = true;
   console.log("hide client successfully.");
};

LAUNCHAPP.prototype.setupForWs1orSDKMode = function () {
   return new Promise((resolve, reject) => {
      var self = this;
      if (!self.mainWindow || !self.chromeClient) {
         return;
      }
      chrome.contextMenus.create({
         title: chrome.i18n.getMessage("CLIENT_SETTINGS_T"),
         id: "settingForWS1",
         contexts: ["launcher"]
      });
      chrome.app.window.create(
         self.shelfPage,
         {
            outerBounds: {
               left: -10000,
               top: -10000,
               width: 1,
               height: 1
            },
            id: "clientMockShelf",
            focused: true,
            frame: "none"
         },
         function (appWindow) {
            self.shelfPlaceHolderWindow = appWindow;
            // Add a listener to detect whether appWindow is closed.
            self.shelfPlaceHolderWindow.onClosed.addListener(function () {
               self.shelfPlaceHolderWindow = null;
               self.mainWindow.close();
            });
            resolve();
         }
      );
      showClientAsSetting = function () {
         self.showClientMainWindow();
         if (typeof self.chromeClient.openSettings === "function") {
            self.chromeClient.openSettings();
         }
      };
   });
};

LAUNCHAPP.prototype.tearDownForURL = function () {
   var self = this;
   self.chromeClient.onWS1ModeInited = undefined;
   self.chromeClient.onSDKModeInited = undefined;
   self.chromeClient.isFromURL = undefined;
   if (self.ws1ModeEnabled || self.SDKModeEnabled) {
      self.tearDownForWs1orSDKMode();
   }
   self.ws1ModeEnabled = null;
   self.SDKModeEnabled = null;
};

LAUNCHAPP.prototype.tearDownForWs1orSDKMode = function () {
   var self = this;
   if (!self.mainWindow || !self.chromeClient) {
      return;
   }
   if (self.shelfPlaceHolderWindow) {
      self.shelfPlaceHolderWindow.close();
      self.shelfPlaceHolderWindow = null;
   }
   self.chromeClient.onSettingClosed = undefined;
   self.chromeClient.showClient = undefined;
   self.chromeClient.hideClient = undefined;
   self.chromeClient.isFromURL = undefined;
   self.chromeClient.isClientHidden = undefined;
   showClientAsSetting = null;
};

// re-enterable with chromeClient, we could use this to clear up after client closed or trigger exist
LAUNCHAPP.prototype.exitClient = function (enforceQuit) {
   // do not reuse Broker session among client instances
   // TODO: using flag in main.js for session control, keep this unchanged for now
   if (this.chromeClient && typeof this.chromeClient.exitClient === "function") {
      console.log("exit client with logout");
      this.chromeClient.exitClient(true);
   } else if (enforceQuit && this.mainWindow) {
      this.mainWindow.close();
      console.log("exit client without logout");
   } else {
      console.log("no client to exit");
   }
   // no matter whether the exist client executed, change status due to we need to Exist Client
   this.skipClearForWaiting = true;
};

LAUNCHAPP.prototype.createFileAssociationMenu = function (fileInfo) {
   var self = this;
   let extensionName = fileInfo.entry.name.match(/\.[^\.]+$/)[0];
   let storageData = {};
   chrome.storage.local.get("fileAssociationHorizon", (res) => {
      if (!res.fileAssociationHorizon) {
         fileInfo.availableApp = {};
      } else {
         storageData = JSON.parse(res.fileAssociationHorizon);
         fileInfo.availableApp = storageData[extensionName];
         fileInfo.extensionName = extensionName;
      }

      chrome.system.display.getInfo((displays) => {
         let dp = displays[0];
         setTimeout(() => {
            chrome.app.window.create(
               self.fileAssociationWindow,
               {
                  id: "HTMLAccessChromeFAWindow",
                  bounds: {
                     width: dp.bounds.width,
                     height: dp.bounds.height
                  },
                  frame: "none",
                  minWidth: 400,
                  minHeight: 370,
                  maxWidth: 400,
                  maxHeight: 370,
                  hidden: false,
                  top: parseInt((dp.bounds.height - 400) / 2),
                  left: parseInt((dp.bounds.width - 400) / 2)
               },
               (appWindow) => {
                  if (!appWindow) {
                     return;
                  }
                  self.faWindow = null;
                  self.mainWindow.contentWindow.fileInfo = fileInfo;
                  appWindow.contentWindow.fileInfo = fileInfo;
                  appWindow.restore();
                  var windowLeft = parseInt((dp.bounds.width - 400) / 2);
                  var windowTop = parseInt((dp.bounds.height - 400) / 2);
                  appWindow.outerBounds.setPosition(windowLeft, windowTop);
                  self.shelfPlaceHolderWindow = appWindow;
                  self.chromeClient.isFAwithMainClientHide = self.isFALaunchWithMainWindowHide;
                  if (self.isFALaunchWithMainWindowHide) {
                     self.mainWindow.minimize();
                  }
                  appWindow.contentWindow.addEventListener("load", () => {
                     self.faWindow = appWindow;
                     console.log("FA window is opened.");
                  });
                  appWindow.onClosed.addListener(() => {
                     self.fileInfo = {};
                     self.faWindow = null;
                  });
               }
            );
            showClientAsSetting = function () {
               self.showClientMainWindow();
               if (typeof self.chromeClient.openSettings === "function") {
                  self.chromeClient.openSettings();
               }
            };
         });
      });
   });
};

/**
 * 1. We always hide the client for URI launch until we know whether the WS1 mode is enabled.
 * 2. for 20fq2 ECR version, no client or broker session is reused
 * 3. for 20fq2 ECR version, no multi-session is supported
 */
LAUNCHAPP.prototype.create = function (index, isFromURL, launchDelay, options, FAInfo) {
   var self = this;
   if (this.mainWindow) {
      if (this.shelfPlaceHolderWindow && this.shelfPlaceHolderWindow.id === "HTMLAccessChromeFAWindow") {
         console.log("Reuse client if chrome client is opened when file association already exists");
         this.mainWindow.restore();
         return;
      } else {
         this.mainWindow.close();
         this.mainWindow = null;
      }
   }
   if (this.shelfPlaceHolderWindow) {
      this.shelfPlaceHolderWindow.close();
      this.shelfPlaceHolderWindow = null;
   }
   this.skipClearForWaiting = false;

   console.log("creating client window");

   var setUpSessionManagementCenter = () => {
      chrome.contextMenus.removeAll();
      chrome.contextMenus.create({
         title: chrome.i18n.getMessage("SESSION_MANAGEMENT_CENTER_T"),
         id: "sessionManagementCenter",
         contexts: ["launcher"]
      });
   };
   var tearDownRightClickMenu = function () {
      chrome.contextMenus.removeAll();
   };

   chrome.system.display.getInfo(function (displays) {
      var dp = displays[0];
      if (!launchDelay) {
         launchDelay = 0;
      }
      if (FAInfo) {
         boundsWidth = 400;
         boundsHeight = 370;
         minWidth = 400;
         minHeight = 370;
         maxWidth = 400;
         maxHeight = 370;
         top = parseInt((dp.bounds.height - 400) / 2);
         left = parseInt((dp.bounds.width - 400) / 2);
         isHidden = false;
      } else {
         boundsWidth = parseInt(dp.bounds.width);
         boundsHeight = parseInt(dp.bounds.height);
         minWidth = 1200;
         minHeight = 900;
         isHidden = isFromURL;
      }
      setTimeout(() => {
         chrome.app.window.create(
            index,
            {
               id: "HTMLAccessChromeWindow",
               bounds: {
                  width: dp.bounds.width,
                  height: dp.bounds.height
               },
               frame: "none",
               minWidth: 1200,
               minHeight: 900,
               hidden: isFromURL
            },
            function (appWindow) {
               if (!appWindow) {
                  return;
               }
               if (FAInfo) {
                  appWindow.restore();
                  var windowLeft = parseInt((dp.bounds.width - 400) / 2);
                  var windowTop = parseInt((dp.bounds.height - 400) / 2);
                  appWindow.outerBounds.setPosition(windowLeft, windowTop);
               }
               self.mainWindow = appWindow;
               setUpSessionManagementCenter();

               appWindow.contentWindow.addEventListener("load", async function () {
                  if (Object.keys(self.fileInfo).length > 0) {
                     self.createFileAssociationMenu(self.fileInfo);
                  }

                  if (self.sdk && typeof self.sdk.onClientLoaded === "function") {
                     self.sdk.onClientLoaded(appWindow, self.sdkClientId, self.debugOptions);
                  }
                  // Set query model in chromeClient after load completely.
                  if (!appWindow.contentWindow.chromeClient) {
                     appWindow.contentWindow.chromeClient = {
                        queryModel: {},
                        useRTAVChannel: false
                     };
                  }

                  self.chromeClient = appWindow.contentWindow.chromeClient;
                  self.chromeClient.isFromURL = isFromURL;
                  self.chromeClient.isFAwithMainClientHide = self.isFALaunchWithMainWindowHide;
                  // always use give client the method to check whether has policy, for user-trust UI
                  self.chromeClient.getTrustPolicy = self.getTrustPolicy.bind(self);

                  if (self.sdkClientId) {
                     self.chromeClient.sdkAssignedId = self.sdkClientId;
                     self.chromeClient.sdkParentAppId = self.sdk.getCurrentInvokerId();
                     let allowNonSAML = self.sdk.debugOption ? self.sdk.debugOption.allowNonSAMLAuth : false;
                     let hideClient = self.launchSettings ? !self.launchSettings.showClient : true;
                     let disableAppSessionAutoResume = self.launchSettings
                        ? self.launchSettings.disableAppSessionAutoResume
                        : true;
                     let showError = self.launchSettings ? self.launchSettings.showError : false;
                     self.chromeClient.sdkOptions = {
                        allowNonSAMLAuth: allowNonSAML,
                        hideClient: hideClient,
                        disableAppSessionAutoResume: disableAppSessionAutoResume,
                        vChanOptions: options.vChanOptions,
                        authSecret: options.authSecret,
                        showError: showError,
                        crypto: options.crypto
                     };
                     self.chromeClient.workingVersion = options.workingVersion;
                  }

                  appWindow.contentWindow.isKioskSession = self.isKioskSession;
                  if (isFromURL && self.url !== null) {
                     self.chromeClient.queryModel.uri = await self.getModelFromURI(decodeURIComponent(self.url));
                  } else {
                     self.chromeClient.queryModel.uri = null;
                  }
                  self.chromeClient.queryModel.policy = {};

                  /**
                   * don't apply the same function for policy for now
                   * Only apply to URI to reduce testing efforts since it's a
                   * 20fq2 ECR
                   */
                  if (isFromURL) {
                     self.setupForURL();
                  }

                  self.chromeClient.clearActiveClient = self.clearActiveClient.bind(self);
                  self.chromeClient.setActiveClient = self.setActiveClient.bind(self);
               });

               // Add a listener to detect whether appWindow is closed.
               appWindow.onClosed.addListener(function () {
                  self.isLaunch = false;
                  self.logoutActiveClient();
                  self.clearActiveClient("client closed initialied reset");
                  tearDownRightClickMenu();
                  if (isFromURL) {
                     self.tearDownForURL();
                  }
                  //release keep-awake in power setting when app is closed
                  if (!!chrome && !!chrome.power) {
                     try {
                        chrome.power.releaseKeepAwake();
                     } catch (e) {}
                  }
                  // Don't call exitClient here again, since it's too late and can't success,
                  // For UI close, it's done in UI event, or URL launch it's done in url event handling
                  // close all opened window after client window is closed
                  chrome.app.window.getAll().forEach((window) => {
                     window.close();
                  });
                  self.chromeClient = null;
                  self.mainWindow = null;
                  if (self.sdk && typeof self.sdk.onClientClosed === "function") {
                     self.sdk.onClientClosed(self.sdkClientId);
                  }
                  self.sdkClientId = null;
               });

               appWindow.onMinimized.addListener(function () {
                  self.mainWindow.contentWindow.isMinimized = true;
               });

               appWindow.onRestored.addListener(function () {
                  self.mainWindow.contentWindow.isMinimized = false;
               });
            }
         );
      }, launchDelay);
   });
};

LAUNCHAPP.prototype.getModelFromURI = async function (uri) {
   var model = {},
      queryReg = uri.split("data=https://").pop(),
      queryURI,
      queryArray,
      queryArrayLength,
      equalLoc;

   // if reg failed, return empty model
   if (!queryReg) {
      return model;
   }
   queryURI = queryReg;
   host = this.getBrokerFromQueryURI(queryURI);

   // If query URI only contains broker address
   if (queryURI.length === host.length) {
      return model;
   }
   queryURI = queryURI.substring(host.length);

   var originServer = await this.getOriginServer(host);
   if (originServer) {
      console.log(host + " is stored redirected from " + originServer);
      model["host"] = originServer;
      queryURI = queryURI + "&mid=" + host;
      queryReg = queryReg.replace(host, originServer) + "&mid=" + host;
   } else {
      model["host"] = host;
   }

   // Remove left ? or /
   while (queryURI.length > 0) {
      if (queryURI[0] === "/" || queryURI[0] === "?") {
         queryURI = queryURI.substring(1);
      } else {
         break;
      }
   }
   let urlParams = queryURI.split("?");
   if (urlParams.length > 2) {
      return model;
   } else if (urlParams.length === 2) {
      queryURI = urlParams[1];
   }

   // Create model data from query array
   queryArray = queryURI.split("&");
   queryArrayLength = queryArray.length;
   for (var i = 0; i < queryArrayLength; i++) {
      // Get the first = location
      equalLoc = queryArray[i].indexOf("=");
      // ignore the url parameter without = and still try to parse others
      if (equalLoc === -1) {
         continue;
      }
      model[queryArray[i].substring(0, equalLoc)] = decodeURIComponent(queryArray[i].substring(equalLoc + 1));
   }
   model.url = queryReg;
   return model;
};

LAUNCHAPP.prototype.getBrokerFromQueryURI = function (queryURI) {
   var address = "",
      queryURILength = queryURI.length;

   for (var i = 0; i < queryURILength; i++) {
      // Broker address may end with ? or /
      if (queryURI[i] === "/" || queryURI[i] === "?") {
         break;
      } else {
         address += queryURI[i];
      }
   }
   return address;
};

LAUNCHAPP.prototype.getHideServerPolicy = function () {
   return new Promise((resolve, reject) => {
      chrome.storage.managed.get((items) => {
         if (
            items &&
            items["broker_list"] &&
            items["broker_list"]["settings"] &&
            items["broker_list"]["settings"]["common-setting"] &&
            items["broker_list"]["settings"]["common-setting"]["hideSecondaryServer"] === false
         ) {
            resolve(false);
         } else {
            resolve(true);
         }
      });
   });
};
/**
 * Check if current server is stored as a secondary server in local storage
 * and return the origin server
 */
LAUNCHAPP.prototype.getOriginServer = function (server) {
   return new Promise((resolve, reject) => {
      this.getHideServerPolicy().then((policy) => {
         if (policy) {
            tempServers = [];
            chrome.storage.local.get("REGISTERED_SERVERS_KEY", (serversObj) => {
               var servers = serversObj["REGISTERED_SERVERS_KEY"];
               if (!servers) {
                  resolve("");
                  return;
               }
               for (var i = 0; i < servers.length; i++) {
                  if (servers[i].secondaryServers) {
                     for (var j = 0; j < servers[i].secondaryServers.length; j++) {
                        if (server === servers[i].secondaryServers[j]) {
                           tempServers.unshift(servers[i]);
                        }
                     }
                  }
               }
               if (tempServers.length === 0) {
                  resolve("");
                  return;
               } else {
                  // Return the most recent server.
                  tempServers.sort((a, b) => {
                     return b.updateTime > a.updateTime ? 1 : -1;
                  });
                  var originServer = tempServers[0].address;
                  resolve(originServer);
                  return;
               }
            });
         } else {
            resolve("");
            return;
         }
      });
   });
};

/**
 * Parse the uri and get its parameters
 */
LAUNCHAPP.prototype.parseURI = function (uri) {
   const a = document.createElement("a");
   a.setAttribute("href", uri);

   let search = a.search || "";
   if (search.startsWith("?")) {
      search = search.substring(1);
   }

   const params = {};
   if (search) {
      search
         .split("&")
         .map((kv) => kv.split("=").map(decodeURIComponent))
         .forEach((kv) => (params[kv[0]] = kv[1]));
   }

   return params;
};

LAUNCHAPP.prototype.getTrustModelFromPolicy = function (policy) {
   const externalTrustKey = "external-trust";
   const trustedOriginsKey = "trusted-list";
   const overwriteDefaultKey = "overwrite-default";
   if (!policy || !policy[externalTrustKey]) {
      return null;
   }
   let model = {
      trustedOrigins: policy[externalTrustKey][trustedOriginsKey] || [],
      overwriteDefault: policy[externalTrustKey][overwriteDefaultKey]
   };
   return model;
};

LAUNCHAPP.prototype.getTrustPolicy = function () {
   return new Promise((resolve, reject) => {
      chrome.storage.managed.get((policy) => {
         let model = this.getTrustModelFromPolicy(policy);
         resolve(model);
      });
   });
};

LAUNCHAPP.prototype.start = function () {
   var self = this;

   var onClientLaunched = async function (event, onBeforeLaunch, options, isSDKLaunched) {
      console.log("detect launch event");
      // See detail in the https://developer.chrome.com/apps/manifest/kiosk_enabled
      self.isKioskSession = !!event.isKioskSession;
      self.url = null;
      var isFromURL = false;
      var launchDelay = 0;
      var clearWaitTime = 0;
      // If invoked by URI.
      if (event.id === "horizon_URL_handler") {
         self.url = event.url;
         isFromURL = true;
         // give time for session to die with previous client to reduce impact of bug 2717716
         if (self.mainWindow) {
            clearWaitTime = 1000; // allow JS execution, which normally takes << 1s
            launchDelay = 1000; // allow OS event trigger, also check onContextMenuClicked
         }
         console.log("URL launch detected" + self.url);
         let urlParams = await self.getModelFromURI(decodeURIComponent(self.url));
         if (self.doURLAction(urlParams, options, isSDKLaunched)) {
            console.log("reuse existing client for processing url");
            return;
         }
      } else if (event.source === "file_handler") {
         isFromURL = true;
         console.log("this is invoked by file association");
         if (event.items) {
            self.fileInfo.entry = event.items[0].entry;
            self.fileInfo.fileType = event.items[0].type;
         } else {
            console.log("normal launch, source: " + event.source);
         }

         if (self.chromeClient) {
            console.log("Client is already opened");
            self.isFALaunchWithMainWindowHide = self.mainWindow.contentWindow.isMinimized;

            if (self.faWindow) {
               //if the FA menu is active, update the info on it. bring it to top
               let extensionName = self.fileInfo.entry.name.match(/\.[^\.]+$/)[0];
               let storageData = {};
               chrome.storage.local.get("fileAssociationHorizon", (res) => {
                  if (!res.fileAssociationHorizon) {
                     self.fileInfo.availableApp = {};
                  } else {
                     storageData = JSON.parse(res.fileAssociationHorizon);
                     self.fileInfo.availableApp = storageData[extensionName];
                     self.fileInfo.extensionName = extensionName;
                  }
               });
               self.mainWindow.contentWindow.fileInfo = self.fileInfo;
               self.faWindow.contentWindow.fileInfo = self.fileInfo;
               //need to update
               let message = {
                  type: "updateFileInfo"
               };
               chrome.runtime.sendMessage(message);
               self.faWindow.focus();
            } else {
               //if the FA menu is inactive, other page is active, don't close other page, create the fa menu with new info
               self.createFileAssociationMenu(self.fileInfo);
            }
            return;
         } else {
            self.isFALaunchWithMainWindowHide = true;
         }
      }

      /**
       * Exit client first otherwise OS would queue the function execution of previous client and do it inside next instance.
       * See bug 2725314 and 2717716.
       */
      if (clearWaitTime) {
         console.log("delay " + clearWaitTime + "s before next launch");
         self.exitClient();
         setTimeout(() => {
            if (typeof onBeforeLaunch === "function") {
               onBeforeLaunch();
            }
            self.create(self.index, isFromURL, launchDelay, options);
         }, clearWaitTime);
      } else {
         console.log("no delay before next launch");
         if (typeof onBeforeLaunch === "function") {
            onBeforeLaunch();
         }
         self.create(self.index, isFromURL, launchDelay, options);
      }
   };
   self.onClientLaunched = onClientLaunched;

   var sessionManagementCenter = null;
   var createSessionManagementCenter = () => {
      var screenWidth = screen.availWidth;
      var screenHeight = screen.availHeight;
      var width = 820;
      var height = 600;
      chrome.app.window.create(
         "./webclient/session-management-center.html",
         {
            id: "sessionManagementCenter",
            outerBounds: {
               minWidth: width,
               minHeight: height,
               maxWidth: width,
               maxHeight: height,
               left: Math.round((screenWidth - width) / 2),
               top: Math.round((screenHeight - height) / 2)
            }
         },
         (app) => {
            if (app) {
               app.contentWindow.isKioskSession = self.isKioskSession;
               sessionManagementCenter = app;
            } else {
               console.log("failed to create window with error ", chrome.runtime.lastError);
            }
         }
      );
   };

   var onContextMenuClicked = function (itemData) {
      if (itemData.menuItemId === "sessionManagementCenter") {
         if (sessionManagementCenter) {
            sessionManagementCenter.close();
            sessionManagementCenter = null;
            setTimeout(() => {
               createSessionManagementCenter();
            }, 1000);
         } else {
            createSessionManagementCenter();
         }
      } else if (itemData.menuItemId === "settingForWS1") {
         if (showClientAsSetting) {
            showClientAsSetting();
         }
      }
   };

   // Add launch listener
   chrome.app.runtime.onLaunched.addListener(onClientLaunched);
   /**
    * remove the chrome.app.runtime.onRestarted to avoid dup invocation, see doc as below:
    * Fired at Chrome startup to apps that were running when Chrome last shut down,
    * or when apps have been requested to restart from their previous state for other reasons
    * (e.g. when the user revokes access to an app's retained files the runtime will restart the app).
    * In these situations if apps do not have an onRestarted handler they will be sent an onLaunched event instead.
    */
   chrome.contextMenus.onClicked.addListener(onContextMenuClicked);
};

let previousTrustChangedCallback = null;
LAUNCHAPP.prototype.onPolicyChanged = (changes, area) => {
   let previousPolicyChangedObject = null;
   const policyKey = "external-trust";
   const storageKey = "TRUST_LIST";
   if (area === "managed" && changes.hasOwnProperty(policyKey)) {
      previousPolicyChangedObject = changes[policyKey];
   } else if (area === "local" && changes.hasOwnProperty(storageKey)) {
      previousPolicyChangedObject = changes[storageKey];
   }
   if (!previousPolicyChangedObject) {
      return;
   }
   if (typeof previousTrustChangedCallback === "function") {
      previousTrustChangedCallback(previousPolicyChangedObject.newValue, previousPolicyChangedObject.oldValue);
   }
};
LAUNCHAPP.prototype.onTrustModelChanged = function (callback) {
   previousTrustChangedCallback = callback;
};

LAUNCHAPP.prototype.getTrustedModel = function () {
   return new Promise((resolve, reject) => {
      const defaultTrustedAppsString = "chrome-extension://cokoeepjbmmnhgdhlkpahohdaiedfjgn, chrome-extension://omificdfgpipkkpdhbjmefgfgbppehke, chrome-extension://plpogimmgnkkiflhpidbibfmgpkaofec, chrome-extension://pllbepacblmgialkkpcceohmjakafnbb, chrome-extension://dahgfgiifpnaoajmloofonkndaaafacp, chrome-extension://egfpnfjeaopimgpiioeedbpmojdapaip, chrome-extension://baobpecgllpajfeojepgedjdlnlfffde, chrome-extension://jcnflhjcfjkplgkcinikhbgbhfldkadl, chrome-extension://jkfjfbelolphkjckiolfcakgalloegek, chrome-extension://fnbibocngjnefolmcodjkkghijpdlnfm".replaceAll(" ", "");
      const vmwareTrustedOrigins = defaultTrustedAppsString ? defaultTrustedAppsString.split(",") : [];
      this.getTrustPolicy()
         .then((trustPolicy) => {
            if (!trustPolicy) {
               console.log("no policy for app trusting found, return default", vmwareTrustedOrigins);
               resolve({
                  hasPolicy: false,
                  defaultTrustList: vmwareTrustedOrigins
               });
               return;
            }

            // not use | to avoid getting boolean
            let policyTrustOrigins = trustPolicy.trustedOrigins;
            if (!policyTrustOrigins) {
               policyTrustOrigins = [];
            }
            if (trustPolicy.overwriteDefault) {
               console.log("policy for app trusting found, overwrite enabled", policyTrustOrigins);
               resolve({
                  hasPolicy: true,
                  defaultTrustList: policyTrustOrigins
               });
            } else {
               let mergedList = vmwareTrustedOrigins;
               if (!!policyTrustOrigins && policyTrustOrigins.length > 0) {
                  policyTrustOrigins.forEach((origin) => {
                     if (mergedList.indexOf(origin) === -1) {
                        mergedList.push(origin);
                     }
                  });
               }
               console.log("policy for app trusting found, merge with default", mergedList);
               resolve({
                  hasPolicy: true,
                  defaultTrustList: mergedList
               });
            }
         })
         .catch(reject);
   });
};

LAUNCHAPP.prototype.registerSDK = function () {
   let sdkEntry = new SDKBackgroundEntry(this.getTrustedModel.bind(this), this.onTrustModelChanged.bind(this));
   sdkEntry.setOperations({
      // reuse URL launch for now
      launchClient: (parameter) => {
         let nextClientId = this.sdk.getNextClientId();
         let allowNonSAML = this.sdk.debugOption ? this.sdk.debugOption.allowNonSAMLAuth : false;
         this.launchSettings = this.sdk.launchSettings;
         // TODO: protect with crypto
         let authSecret = parameter.authInfo ? parameter.authInfo.password : undefined;
         let url = this.convertSDKLaunchToURL(parameter, allowNonSAML);
         if (url) {
            let mockedEvent = {
               id: "horizon_URL_handler",
               url: url
            };
            console.log("launch for SDK by URL " + url);
            this.onClientLaunched(
               mockedEvent,
               () => {
                  this.sdkClientId = nextClientId;
               },
               {
                  vChanOptions: parameter.vChanOptions,
                  authSecret: authSecret,
                  crypto: parameter.crypto,
                  workingVersion: parameter.workingVersion
               },
               true
            );
         } else {
            console.error("not supported launch from SDK");
            this.sdk.onClientLaunchFailed("invalid parameter found when processing request, please check input format");
         }
      },
      getClient: () => {
         return this.mainWindow;
      },
      closeClient: () => {
         this.exitClient(true);
      },
      showClient: this.showClientMainWindow.bind(this),
      hideClient: this.hideClientMainWindow.bind(this),
      getClientVisibility: () => {
         return !!this.chromeClient && !this.chromeClient.isClientHidden;
      }
   });
   chrome.storage.onChanged.addListener(this.onPolicyChanged.bind(this));
   this.sdk = sdkEntry;
};

//TODO
LAUNCHAPP.prototype.convertSDKLaunchToURL = (launchingParameters, allowNonSAML) => {
   console.log(launchingParameters);
   let parameters = [];
   if (launchingParameters.type === "HorizonDesktop") {
      parameters.push("desktopId=" + encodeURIComponent(launchingParameters.id));
   } else if (launchingParameters.type === "HorizonApp") {
      parameters.push("applicationId=" + encodeURIComponent(launchingParameters.id));
   } else if (launchingParameters.type === "HorizonAppSession") {
      parameters.push("itemId=" + encodeURIComponent(launchingParameters.id));
   } else {
      return null;
   }
   if (launchingParameters.authInfo.hasOwnProperty("samlArtifact") && !!launchingParameters.authInfo.samlArtifact) {
      parameters.push("SAMLart=" + encodeURIComponent(launchingParameters.authInfo.samlArtifact));
   } else if (!allowNonSAML) {
      console.error("not allow non-SAML auth");
      return null;
      // For v1, simplized flow require both user name and password for launch from SDK
   } else if (
      launchingParameters.authInfo.hasOwnProperty("username") &&
      !!launchingParameters.authInfo.username &&
      launchingParameters.authInfo.hasOwnProperty("password") &&
      !!launchingParameters.authInfo.password
   ) {
      parameters.push("user=" + encodeURIComponent(launchingParameters.authInfo.username));
      if (launchingParameters.authInfo.hasOwnProperty("domain") && !!launchingParameters.authInfo.domain) {
         parameters.push("domainName=" + encodeURIComponent(launchingParameters.authInfo.domain));
      }
   }
   if (launchingParameters.serverId) {
      parameters.push("serverId=" + encodeURIComponent(launchingParameters.serverId));
   }
   let parameterStrings = parameters.join("&");
   if (parameterStrings) {
      parameterStrings = "?" + parameterStrings;
   }
   let brokerAddress = launchingParameters.serverAddress;
   return "SDKLauncher?data=" + encodeURIComponent("https://" + brokerAddress + parameterStrings);
};

/**
 * Add listeners to initialize URL redirection's rules
 */
LAUNCHAPP.prototype.registerUrlRedirection = function () {
   // Update rules when installing the client
   chrome.runtime.onInstalled.addListener((details) => {
      if (details.reason === chrome.runtime.OnInstalledReason.INSTALL) {
         console.log("Update the URL redirection rule when installing the client");
         this.updateUrlRedirectRule();
      }
   });
   // Update rules when google policy's rule changes
   chrome.storage.managed.onChanged.addListener((changes) => {
      if (changes && changes["broker_list"]) {
         chrome.storage.local.get(URL_REDIRECTION_RULE_SOURCE_KEY, (items) => {
            // If the rules are fetched from the broker, cannot be overridden with google policy's rules
            if (items && items[URL_REDIRECTION_RULE_SOURCE_KEY] !== "broker") {
               console.log("Update the URL Redirection rules when google policy changes");
               this.updateUrlRedirectRule();
            } else {
               console.log(
                  "Cannot update the URL Redirection rules when google policy changes, because the existing rules come from the broker"
               );
            }
         });
      }
   });
   // Update the rules when the chrome extension needs to initialize the rules
   chrome.runtime.onMessageExternal.addListener((request, sender) => {
      if (sender.id === URL_REDIRECTION_EXTENSION_ID && request && request.type === "getRule") {
         console.log("Received getRule message from the URL redirection chrome extension");
         this.updateUrlRedirectRule();
      }
   });
};

/**
 * Fetch the URL redirection rule from google policy and send to the chrome extension
 */
LAUNCHAPP.prototype.updateUrlRedirectRule = function () {
   chrome.storage.managed.get((items) => {
      const msg = {
         type: "updateRule",
         content: {
            enabled: true,
            host: "chrome",
            origin: "client",
            protocols: []
         }
      };
      if (
         items &&
         items["broker_list"] &&
         items["broker_list"]["settings"] &&
         items["broker_list"]["settings"]["common-setting"] &&
         items["broker_list"]["settings"]["common-setting"]["urlFiltering"] &&
         items["broker_list"]["settings"]["common-setting"]["urlFiltering"]["protocols"] &&
         items["broker_list"]["settings"]["common-setting"]["urlFiltering"]["protocols"].length
      ) {
         msg.content.protocols = items["broker_list"]["settings"]["common-setting"]["urlFiltering"]["protocols"];
      } else {
         console.log("There is no URL redirection rule in google policy");
      }
      console.log("Send the message to chrome extension to update URL redirection rule");
      chrome.runtime.sendMessage(URL_REDIRECTION_EXTENSION_ID, msg, null, () => {
         // Storage redirection rules is fetch from google policy, used to determine to synchronize redirection rules when google policy changes
         chrome.storage.local.set({ [URL_REDIRECTION_RULE_SOURCE_KEY]: "google_policy" });
      });
   });
};

var launchApp = new LAUNCHAPP();
console.info("The client background page is reloaded");
try {
   launchApp.registerSDK();
   launchApp.registerUrlRedirection();
} catch (e) {
   console.error(e);
}
launchApp.start();
